/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
# include "uart2.h"

void uart_tx_string ( uint8_t *s )
{
    while ( *s )
        uart_tx_char ( *s++ ) ;
}

int main ( void )
{
    uint8_t rc ;
    int c_pos = FIRST_POSITION ; 
    int c_line = FIRST_LINE ;

    uart_init ( BAUD_RATE ) ;
    uart_tx_string ( "Enter input: " ) ;
    lcd_clrscr( ) ;

    for ( ; ; )
    {
        /* cursor  */
        lcd_gotoxy ( c_pos , c_line ) ;
        lcd_putc ( '_' ) ;

        rc = uart_rx_char( ) ;

        switch ( rc )
        {
            case FRAME_ERROR :
            case DATA_OVERRUN_ERROR :
            case PARITY_ERROR :
                lcd_puts_p ( PSTR ( "ERR" ) ) ;
                break ;

            case BACKSPACE_KEY :
            case DELETE_KEY :
                lcd_gotoxy ( c_pos, c_line ) ;
                lcd_putc ( ' ' ) ;

                // if cursor is in second line and first position
                if ( c_line == SECOND_LINE &&
                        c_pos == FIRST_POSITION )
                {
                    c_line = FIRST_LINE ;
                    c_pos  = LAST_POSITION ;
                }

                if ( c_pos != FIRST_POSITION )
                {
                    // move to character that has to be deleted
                    c_pos-- ;
                    lcd_gotoxy ( c_pos , c_line ) ;
                    // Delete the character
                    lcd_putc ( ' ' ) ;
                }

                lcd_gotoxy ( c_pos , c_line ) ;
                uart_tx_char ( '\b' ) ;
                break ;

            case ENTER_KEY :

                if ( c_line == FIRST_LINE )
                {
                    lcd_gotoxy ( c_pos , c_line ) ;
                    lcd_putc ( ' ' ) ;

                    c_line = SECOND_LINE ;
                    c_pos = FIRST_POSITION ;
                }

                uart_tx_char ( '\r' ) ;
                break ;

            default :

                lcd_gotoxy ( c_pos , c_line ) ;            
                if ( rc == '\r' )
                    rc = '\n' ;
                else if ( rc == '\t' )
                    rc = ' ' ;
                lcd_putc ( rc ) ;

                // Increment x if LAST_POSITION (End of line) 
                if ( c_pos != LAST_POSITION )
                    c_pos ++ ;

                if ( c_pos == LAST_POSITION &&  
                    c_line == FIRST_LINE )
                {
                    // find if reach the end of first line if so
                    // move to first position of the second line
                    c_line = SECOND_LINE ;
                    c_pos = FIRST_POSITION ;
                }
                break ;
        }
    }
    return 0 ;
}
