/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
# include "uart.h"

# define UART_BAUD_SELECT   ( F_CPU / ( baud * 16l ) - 1 )

volatile    static INITFUNC_T uart_InitFunc ;
volatile    static RXFUNC_T uart_RxFunc ;
volatile    static TXFUNC_T uart_TxFunc ;

void uart_init ( unsigned int baud, uint8_t enable_ints )
{
    UBRRH = UART_BAUD_SELECT >> 8 ;
    UBRRL = ( uint8_t ) UART_BAUD_SELECT ;

    if ( enable_ints )
        UCSRB = _BV( RXEN ) | _BV ( TXEN ) | _BV( RXCIE ) | 
               _BV( TXCIE ) ;
    else
        UCSRB = _BV( RXEN ) | _BV ( TXEN ) ;
        
    UCSRC = _BV ( URSEL ) | _BV ( UCSZ1 ) | _BV ( UCSZ0 ) ;

    if ( uart_InitFunc )
        uart_InitFunc( ) ;
}

void uart_set_initfunc ( INITFUNC_T init_func )
{
    uart_InitFunc = init_func ;
}

void uart_set_rxfunc ( RXFUNC_T rx_func )
{
    uart_RxFunc = rx_func ;
}

void uart_set_txfunc ( TXFUNC_T tx_func )
{
    uart_TxFunc = tx_func ;
}

void uart_tx_char ( unsigned char ch )
{
    if ( uart_TxFunc )
        uart_TxFunc ( ch ) ;
}

SIGNAL ( SIG_UART_RECV )
{
    uint8_t c = UDR ;
    if ( uart_RxFunc )
        uart_RxFunc ( c ) ;
}

SIGNAL ( SIG_UART_TRANS )
{
    if ( uart_TxFunc )
        uart_TxFunc ( 0xff ) ;
}
