/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
#include "uart1.h"
#include <avr/interrupt.h>
#include "lcd_io.h"
#include "timer.h"

#define UART_BAUD_SELECT   ( F_CPU / ( baud * 16L ) - 1 )

volatile uint8_t    tx_ready ;

void uart_init ( unsigned int baud )
{
    UBRRH = UART_BAUD_SELECT >> 8 ;
    UBRRL = ( uint8_t ) UART_BAUD_SELECT ;

    /* Enable 8-bit receiver and transmitter */
    UCSRB = _BV( RXEN ) | _BV ( TXEN ) | _BV( TXCIE ) ;
    UCSRC = _BV ( URSEL ) | _BV ( UCSZ1 ) | _BV ( UCSZ0 ) ;
    tx_ready = 1 ;
}

void uart_tx_char (unsigned char s )
{
    timer_start( ) ;
    while ( !tx_ready )
        ;
    
    tx_ready = 0 ;
    UDR = s ;
    lcd_putc ( s ) ;
}

SIGNAL ( SIG_UART_TRANS )
{
    uint16_t    t = timer_read_usec ( ) ;
    lcd_clrscr ( ) ;
    lcd_puts( "Time: " ) ;
    lcd_puti( t, 0 ) ;
    tx_ready = 1 ;
}
