/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
#include "pager.h"

static void init_avr ( void ) ;
static void start_timer ( void ) ;

static void kp_init ( volatile struct keypad *kp ) ;
static void kp_display ( volatile struct keypad *kp ) ;
static void kp_left ( volatile struct keypad *kp ) ;
static void kp_right ( volatile struct keypad *kp ) ;
static void kp_show_line1 ( volatile struct keypad *kp ) ;

volatile struct keypad *active_kp ;
volatile struct keypad ascii_kp ;

int main ( void )
{
    PGM_P   alphabets =
        PSTR ( "abcdefghijklmnopqrstuvwxyz0123456789" ) ;

    init_avr( ) ;
    kp_init ( &ascii_kp ) ;

    ascii_kp.keys = alphabets ;
    ascii_kp.num_keys = strlen_P ( alphabets ) ;

    active_kp = &ascii_kp ;
    kp_display ( active_kp ) ;
    return 0 ;
}

void init_avr ( void )
{
    lcd_init( ) ;
    buttons_init( ) ;
    start_timer( ) ;
    lcd_command ( LCD_DISP_ON_BLINK ) ;
    sei( ) ;
}

void start_timer ( void )
{
    TCNT0 = 0x00 ; /* Overflow after 32.768 msec */
    TCCR0 = 0x05 ; /* Divide by 1024 */
    TIMSK = 0x01 ; /* Enable TOV0 */
}

SIGNAL ( SIG_OVERFLOW0 )
{
    buttons_check( ) ;
    if ( left_state == RELEASED )
    {
        RESET_KEY_STATE ( left_state ) ;
        kp_left ( active_kp ) ;
    }
    else if ( right_state == RELEASED ) 
    {
        RESET_KEY_STATE ( right_state ) ;
        kp_right ( active_kp ) ;
    }
    else if ( select_state == RELEASED ) 
    {
        RESET_KEY_STATE ( select_state ) ;
        kp_show_line1 ( active_kp ) ;
    }
}

void kp_init ( volatile struct keypad *kp )
{
    memset( (void *)kp, sizeof ( *kp ), 0 ) ;
}

void kp_display ( volatile struct keypad *kp )
{
    uint8_t     i ;
    uint8_t     cnt = 0 ;

    lcd_gotoxy ( 0 , 1 ) ;
    for ( i = kp->k_pos ; cnt < LCD_CHARS_PER_LINE ; i ++ )
    {
        lcd_putc ( pgm_read_byte ( & kp->keys [ i % kp->num_keys ] ) ) ;
        cnt ++ ;
    }

    kp->x_coord = 0 ;
    lcd_gotoxy ( kp->x_coord , 1 ) ;
}

void kp_left ( volatile struct keypad *kp )
{
    if ( kp->k_pos == 0 )
        kp->k_pos = kp->num_keys - 1 ;
    else
        kp->k_pos -- ;
    if ( kp->x_coord > 0 )
    {
        lcd_gotoxy ( -- kp->x_coord , 1 ) ;
        return ;
    }
    else
        kp_display ( kp ) ;
}

void kp_right ( volatile struct keypad *kp )
{
    if ( kp->x_coord < LCD_CHARS_PER_LINE )
    {
        kp->k_pos ++ ;
        kp->k_pos %= kp->num_keys ;
        lcd_gotoxy ( ++ kp->x_coord , 1 ) ;
        return ;
    }
    else
        kp_display ( kp ) ;
}

void kp_show_line1 ( volatile struct keypad *kp )
{
    lcd_gotoxy ( kp->out_cur ++ , 0 ) ;
    kp->out_cur %= LCD_CHARS_PER_LINE ;
    lcd_putc ( pgm_read_byte ( & kp->keys [ kp->k_pos ] ) ) ;
    lcd_gotoxy( kp->x_coord , 1 ) ;
}

