/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
#include "twi_driver.h"

/* Address of ATmega32 if it is used as 'Slave Receiver' */
# define OWN_ADDR     0x60

static uint8_t _get_byte ( uint8_t*, char not_last_byte ) ;
static void twi_wait_int ( void ) ;

void twi_init ( void )
{
    TWAR = OWN_ADDR ;
    TWBR = ( F_CPU / 100000UL - 16 ) / 2 ;
    TWCR = _BV ( TWEN ) ;
    return ;
}

void twi_wait_int ( void )
{
    while ( ! ( TWCR & _BV( TWINT ) ) )
        ;
}    

uint8_t twi_send_start ( void )
{
    TWCR = _BV ( TWINT ) | _BV ( TWSTA ) | _BV ( TWEN ) ;

    twi_wait_int( ) ;  // Wait for TWI interrupt flag to be set
    if ( ( TWSR != TW_START ) && ( TWSR != TW_REP_START ) )
        return TWSR ;
    return SUCCESS ;
}

void twi_send_stop ( void )
{
    TWCR = _BV ( TWINT ) | _BV ( TWSTO ) | _BV ( TWEN ) ;
}

uint8_t twi_send_byte ( uint8_t data )
{
    twi_wait_int( ) ;

    TWDR = data ;
    TWCR = _BV ( TWINT ) | _BV ( TWEN ) ;

    twi_wait_int( ) ;

    if ( TWSR != TW_MT_DATA_ACK )
        return TWSR ;
    
    return SUCCESS ;
}

/* Send a SLA + W/R to the bus */
uint8_t twi_send_slave_address ( uint8_t adr )
{
    twi_wait_int( ) ;

    TWDR = adr ;
    TWCR = _BV ( TWINT ) | _BV ( TWEN ) ;

    twi_wait_int( ) ;

    if ( TWSR != TW_MT_SLA_ACK && TWSR != TW_MR_SLA_ACK )
        return TWSR ;
    
    return SUCCESS ;
}

uint8_t   _get_byte ( uint8_t *rx_ptr,char not_last_byte )
{
    twi_wait_int( ) ;

    /* When receiving last byte from slave, it will be sent a NACK. 
    This makes the slave stop transmitting. */
    if ( not_last_byte ) 
        // Clear int flag and enable acknowledge to receive data.
        TWCR = _BV ( TWINT ) | _BV ( TWEN ) | _BV ( TWEA ) ;
    else 
        /* Don't enable ACK to tell slave to stop transmitting */
        TWCR = _BV ( TWINT ) | _BV ( TWEN ) ;
 
    twi_wait_int( ) ; 
    *rx_ptr = TWDR ;         // Received byte from slave

    /* If ACK has been transmitted or this was the last byte & NACK has 
    been sent then return SUCCESS */
    if ( ( ( TWSR == TW_MR_DATA_NACK ) && ( not_last_byte == 0 ) ) || 
        ( TWSR == TW_MR_DATA_ACK ) )
        return SUCCESS ;   

    return TWSR ;
}

uint8_t twi_receive_byte ( char not_last_byte )
{
    uint8_t  ch ;
    uint8_t  status = _get_byte ( & ch , not_last_byte ) ;
    if ( status == SUCCESS )
        return ch ;
    else
        return 0xff ;
}
