/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
#include "rtc_write.h"

static uint8_t 	lcd_prompt ( uint8_t max, uint8_t min, PGM_P prompt, uint8_t cur_val ) ;
static void 	start_timer ( void ) ;

uint8_t rtc_write_time ( cTime_t *time )
{
    uint8_t     rc ;

    rc = twi_send_start ( ) ;
    if ( rc == SUCCESS )               
        rc = twi_send_slave_address ( DS1307_ADDR + TW_WRITE ) ;

    if ( rc == SUCCESS )               
        rc = twi_send_byte ( 0x00 ) ;

    if ( rc == SUCCESS )
    {
        twi_send_byte ( time->sec ) ;
        twi_send_byte ( time->min ) ;
        twi_send_byte ( time->hour ) ;
        twi_send_byte ( time->day ) ;
        twi_send_byte ( time->date ) ;
        twi_send_byte ( time->month ) ;
        twi_send_byte ( time->year ) ;
    
        twi_send_stop ( ) ;
    }
    return rc ;
}

uint8_t lcd_prompt ( uint8_t max , uint8_t min , PGM_P prompt , uint8_t cur )
{
    uint8_t val = cur ;

    lcd_clrscr ( ) ;
    while ( 1 ) {
        if (up_state == RELEASED) {
            RESET_KEY_STATE ( up_state ) ;
            if ( val < max )
                val++ ;
            else
                val = min ;
        }
        if (down_state == RELEASED ) {
            RESET_KEY_STATE ( down_state ) ;
            if ( val > min )
                val-- ;
            else 
                val = max;
        }
        lcd_gotoxy ( 0 , 0) ;
        lcd_put_i ( val , 0, 2 ) ;
        lcd_putc ( '\n' ) ;
        lcd_puts_p ( prompt ) ;
        if ( ok_state == RELEASED ) {
            RESET_KEY_STATE( ok_state ) ;
            break ;
        }
        _delay_ms ( 200 ) ;
    }
    val = bin_to_bcd ( val ) ;
    return val ;
}

int main ( void )
{
    cTime_t     tm ;

    buttons_init ( ) ;
    start_timer ( ) ;
    sei ( ) ;
    rtc_init ( ) ;

    rtc_read_time ( &tm ) ;

    tm.sec   = lcd_prompt ( 59, 0, PSTR ("Enter Sec") , tm.sec ) ;
    tm.min   = lcd_prompt ( 59, 0, PSTR ("Enter Min") , tm.min ) ;
    tm.hour  = lcd_prompt ( 23, 0, PSTR ("Enter Hour") , tm.hour ) ;
    tm.day   = lcd_prompt ( 7, 1, PSTR ("Enter Day") , tm.day ) ;
    tm.date  = lcd_prompt ( 31, 1, PSTR ("Enter Date") , tm.date ) ;
    tm.month = lcd_prompt ( 12, 1, PSTR ("Enter Month") , tm.month ) ;
    tm.year  = lcd_prompt ( 20, 0, PSTR ("Enter Year") , tm.year ) ;

    rtc_write_time ( &tm ) ;

    rtc_display_time( ) ;

    return 0;
}

void start_timer()
{
    TCNT0 = 0x00 ;
    TCCR0 = 0x05 ;
    TIMSK = 0x01 ;
}

SIGNAL(SIG_OVERFLOW0)
{
    buttons_check();
}
