#ifndef _IO_REG_MACRO_MODE_
#define _IO_REG_MACRO_MODE_  1     /* In case you have the new assignment mode io headers */
#endif

//#ifndef  _SFR_ASM_COMPAT
//#define  _SFR_ASM_COMPAT     1     /* This is for GCC 3.2 */
//#endif

#if AVRGCC_VERSION >= 330

#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "lcd_io.h"
#if LCD_IO_MODE == 3
#include "i2c.h"
#endif

#else

#include <io.h>
#include <eeprom.h>
#include <pgmspace.h>
#include "lcd_io.h"
#if LCD_IO_MODE == 3
#include "i2c.h"
#endif

#endif /* #if AVRGCC_VERSION == 330 */






#if !defined(outp)
#define outp(value,port)  outb(port,value)
#endif

#if !defined(inp)
#define inp(port)  inb(port)
#endif

/*******************************************************************************************************/
#if defined(__AVR_ATmega103__)

#undef DDRC 
#undef DDRF 
#undef PINC 
#undef PORTF

#if SREG > 0X3F

#define DDRC   0X20
#define DDRF   0X20
#define PINC   0X20
#define PORTF  0X20

#elif SREG <= 0X3F 

#define DDRC   0
#define DDRF   0
#define PINC   0
#define PORTF  0

#endif  /* #if SREG > 0X3F */ 

#endif  /* #if defined(__AVR_ATmega103__) */
/*******************************************************************************************************/


/* constants/macros */
#ifndef CONCAT1
#define CONCAT1(a, b) CONCAT2(a, b)
#endif

#ifndef CONCAT2
#define CONCAT2(a, b) a ## b
#endif


/* Some clever (i hope!) thinking tricks */
#if LCD_IO_MODE == 7 && NUMBER_OF_LCD_UNITS == 1

#warning "MODE 7 USING 1 LCD UNIT. WILL REVERT TO MODE 6"
#undef   LCD_IO_MODE
#define  LCD_IO_MODE            6
#define  LCD_READ_REQUIRED      0
#define  LCD_E_SIGNAL_PORT      LCD_0_E_SIGNAL_PORT
#define  LCD_E_PIN              LCD_0_E_PIN 
#define  LCD_CHARS_PER_LINE     LCD_0_CHARS_PER_LINE
#define  LCD_LINES              LCD_0_LINES 
#undef   LCD_MULTIPLEX_ENABLE
#define  LCD_MULTIPLEX_ENABLE   0
#undef   LCD_READ_REQUIRED
#define  LCD_READ_REQUIRED      0
#undef   LCD_1_CHARS_PER_LINE     
#undef   LCD_1_LINES              
#undef   LCD_1_E_SIGNAL_PORT      
#undef   LCD_1_E_PIN              

           
#elif   LCD_IO_MODE == 7 && NUMBER_OF_LCD_UNITS > 1

#if NUMBER_OF_LCD_UNITS > 3
#error "UP TO 3 LCD UNITS ARE ALLOWED!"
#endif


#endif /* #if LCD_IO_MODE == 7 && NUMBER_OF_LCD_UNITS == 1 */

#if LCD_SAVE_MORE_CODE_SPACE == 1
#warning "LCD IN ABSOLUTELY MINIMUM CONFIGURATION"
#endif

#if LCD_READ_REQUIRED == 1 && LCD_SAVE_MORE_CODE_SPACE == 1
#undef LCD_BACKUP_REQUIRED
#define LCD_BACKUP_REQUIRED   0
#warning "NO BACKUP FUNCTIONS WHILE MINIMUM CODE SELECTED"
#endif 

#if LCD_MULTIPLEX_ENABLE == 1 && LCD_READ_REQUIRED == 1
#undef  LCD_MULTIPLEX_ENABLE 
#define LCD_MULTIPLEX_ENABLE   0
#warning "NO MULTIPLEXING WHILE LCD READING WANTED"
#endif

/* Calculation of the LCD_E_PULSE duration in cpu cycles. sbi() and cbi() are taken in to account. */

#define LCD_E_PULSE_CYCLES_ACCURATE  ( (LCD_E_PULSE_WIDTH_US*1000L)/(1000000000/(F_CPU*10)) )
#define LCD_E_PULSE_CYCLES_REAL      (LCD_E_PULSE_CYCLES_ACCURATE/10)

#if   LCD_E_PULSE_CYCLES_ACCURATE - (LCD_E_PULSE_CYCLES_REAL*10) >=5
#define LCD_E_PULSE_CYCLES           LCD_E_PULSE_CYCLES_REAL
#else 
#define LCD_E_PULSE_CYCLES           (LCD_E_PULSE_CYCLES_REAL-1)
#endif

#if   LCD_E_PULSE_CYCLES <= 0
#define LCD_E_DELAY()             { }

#elif LCD_E_PULSE_CYCLES == 1
#define LCD_E_DELAY()             __asm__ volatile("nop")

#elif LCD_E_PULSE_CYCLES == 2 
#define LCD_E_DELAY()             __asm__ volatile("nop \n\t nop \n\t" ::)

#elif LCD_E_PULSE_CYCLES >= 3


#define DELAY_C_ACCURATE(x)      ({  __asm__ volatile (  "ldi r25,%0 \n\t"           \
                                                         "L_%=:      \n\t"           \
                                                         "dec r25    \n\t"           \
                                                         "brne L_%=  \n\t"           \
                                                         : /* NO OUTPUT */           \
                                                         : "g" ((unsigned char)(x))  \
		                                         : "r25"	             \
           	                                       );                            \
                                  })                          


#define LCD_E_DELAY_LOOPS          (LCD_E_PULSE_CYCLES/3)
  

#if   LCD_E_PULSE_CYCLES-(LCD_E_DELAY_LOOPS*3) == 0

#define  LCD_E_DELAY()            DELAY_C_ACCURATE(LCD_E_DELAY_LOOPS) 

#elif LCD_E_PULSE_CYCLES-(LCD_E_DELAY_LOOPS*3) == 1
 
#define  LCD_E_DELAY()            { DELAY_C_ACCURATE(LCD_E_DELAY_LOOPS); __asm__ volatile("nop"); } 

#elif LCD_E_PULSE_CYCLES-(LCD_E_DELAY_LOOPS*3) == 2
 
#define  LCD_E_DELAY()            { DELAY_C_ACCURATE(LCD_E_DELAY_LOOPS);                         \
                                    __asm__ volatile("nop \n\t nop \n\t" ::);                    \
                                  } 
#elif LCD_E_PULSE_CYCLES-(LCD_E_DELAY_LOOPS*3) == 3
 
#define  LCD_E_DELAY()            { DELAY_C_ACCURATE(LCD_E_DELAY_LOOPS);                         \
                                    __asm__ volatile("nop \n\t nop \n\t nop \n\t" ::);           \
                                  } 
#elif LCD_E_PULSE_CYCLES-(LCD_E_DELAY_LOOPS*3) == 4
 
#define  LCD_E_DELAY()            { DELAY_C_ACCURATE(LCD_E_DELAY_LOOPS);                         \
                                    __asm__ volatile("nop \n\t nop \n\t nop \n\t nop \n\t" ::);  \
                                  } 


#endif 

#endif /* #if   LCD_E_PULSE_CYCLES == 1 */                                 

#define LCD_POWER_ON_DELAY ( (50*(F_CPU/6000))-3 )    /* 50 milliseconds */
#define LCD_INIT_DELAY     ( (10*(F_CPU/6000))-3 )    /* 10 milliseconds */


/* CONVERSION OF LCD_DELAY_TIME_US TO TRUE MICROSECONDS */


#define LCD_SMALL_DELAY    ( ((LCD_DELAY_TIME_US*(F_CPU/60000))/100)-3 )
#define LCD_LONG_DELAY     ( LCD_SMALL_DELAY*40 )





/*######################################################################################################*/
#if LCD_IO_MODE == 2
/*######################################################################################################*/

#define LCD_DATA_OUT_REG      CONCAT1(PORT,LCD_DATA_SIGNAL_PORT)
#define LCD_DATA_DDR_REG      CONCAT1(DDR,LCD_DATA_SIGNAL_PORT)

#define LCD_CLOCK_OUT_REG     CONCAT1(PORT,LCD_CLOCK_SIGNAL_PORT) 
#define LCD_CLOCK_DDR_REG     CONCAT1(DDR,LCD_CLOCK_SIGNAL_PORT) 

#define LCD_READ_REQUIRED     0

#define Toggle_clock()        ({ sbi(LCD_CLOCK_OUT_REG, LCD_CLOCK_PIN); LCD_E_DELAY(); \
                                 cbi(LCD_CLOCK_OUT_REG, LCD_CLOCK_PIN); LCD_E_DELAY(); \
                              }) 

#define DATA(x)               ({ if(x) sbi(LCD_DATA_OUT_REG, LCD_DATA_PIN); \
                                 else cbi(LCD_DATA_OUT_REG, LCD_DATA_PIN);  \
                              })
        
#define Toggle_data()         ({ sbi(LCD_DATA_OUT_REG, LCD_DATA_PIN); LCD_E_DELAY();   \
                                 cbi(LCD_DATA_OUT_REG, LCD_DATA_PIN); LCD_E_DELAY();   \
                              })



#if defined(__AVR_ATmega103__)
#if SREG > 0X3F

#if LCD_DATA_OUT_REG == 0x20 || LCD_CLOCK_OUT_REG == 0x20
#error "SOME OF THE LCD PINS ARE INPUTS ONLY"
#endif

#elif SREG <= 0X3F 
#if LCD_DATA_OUT_REG == 0 || LCD_CLOCK_OUT_REG == 0
#error "SOME OF THE LCD PINS ARE INPUTS ONLY"
#endif

#endif  /* #if SREG > 0X3F */ 

#endif  /* #if defined(__AVR_ATmega103__) */

#define LCD_DONT_READ_BUSY_FLAG  0 
#define LCD_READ_REQUIRED        0
#define NUMBER_OF_LCD_UNITS      1

/*######################################################################################################*/
#elif LCD_IO_MODE == 3
/*######################################################################################################*/



/*######################################################################################################*/
#elif LCD_IO_MODE == 6
/*######################################################################################################*/

#define lcd_toggle_e()  ({ cbi(LCD_E_OUT_REG, LCD_E_PIN); LCD_E_DELAY();  \
                           sbi(LCD_E_OUT_REG, LCD_E_PIN); LCD_E_DELAY();  \
                        })

#if LCD_READ_REQUIRED == 1  
#undef  LCD_MULTIPLEX_REQUIRED
#define LCD_MULTIPLEX_REQUIRED   0
#endif

#if LCD_LINES == 1 || LCD_LINES == 2 || LCD_LINES == 4
#else
#error THE LCD LINES MUST BE 1 OR 2 OR 4 !
#endif



#define LCD_D4_OUT_REG       CONCAT1(PORT, LCD_DATA4_PORT)
#define LCD_D5_OUT_REG       CONCAT1(PORT, LCD_DATA5_PORT)
#define LCD_D6_OUT_REG       CONCAT1(PORT, LCD_DATA6_PORT)
#define LCD_D7_OUT_REG       CONCAT1(PORT, LCD_DATA7_PORT)
#define LCD_RS_OUT_REG       CONCAT1(PORT, LCD_RS_SIGNAL_PORT)
#define LCD_RW_OUT_REG       CONCAT1(PORT, LCD_RW_SIGNAL_PORT)
#define LCD_E_OUT_REG        CONCAT1(PORT, LCD_E_SIGNAL_PORT)

#define LCD_D4_DDR_REG       CONCAT1(DDR, LCD_DATA4_PORT)
#define LCD_D5_DDR_REG       CONCAT1(DDR, LCD_DATA5_PORT)
#define LCD_D6_DDR_REG       CONCAT1(DDR, LCD_DATA6_PORT)
#define LCD_D7_DDR_REG       CONCAT1(DDR, LCD_DATA7_PORT)
#define LCD_RS_DDR_REG       CONCAT1(DDR, LCD_RS_SIGNAL_PORT)
#define LCD_RW_DDR_REG       CONCAT1(DDR, LCD_RW_SIGNAL_PORT)
#define LCD_E_DDR_REG        CONCAT1(DDR, LCD_E_SIGNAL_PORT)



#if   LCD_READ_REQUIRED == 1

#define LCD_D4_PIN_REG        CONCAT1(PIN, LCD_DATA4_PORT)
#define LCD_D5_PIN_REG        CONCAT1(PIN, LCD_DATA5_PORT)
#define LCD_D6_PIN_REG        CONCAT1(PIN, LCD_DATA6_PORT)
#define LCD_D7_PIN_REG        CONCAT1(PIN, LCD_DATA7_PORT)

#elif LCD_READ_REQUIRED == 0

#undef LCD_RW_OUT_REG
#undef LCD_RW_DDR_REG

#endif

#if defined(__AVR_ATmega103__)
#if SREG > 0X3F

#if LCD_D4_OUT_REG == 0x20 || LCD_D5_OUT_REG == 0x20 || LCD_D6_OUT_REG == 0x20 ||  \
    LCD_D7_OUT_REG == 0x20 || LCD_RS_OUT_REG == 0x20 || LCD_E_OUT_REG == 0x20
#error "SOME OF THE LCD PINS ARE INPUTS ONLY"
#endif

#if (LCD_D4_DDR_REG == 0x20 || LCD_D5_DDR_REG == 0x20 || LCD_D6_DDR_REG == 0x20 ||  \
     LCD_D7_DDR_REG == 0x20) && LCD_READ_REQUIRED == 1 
#error "THE LCD DATA PINS ARE NOT BIDIRECTIONAL"
#endif

#elif SREG <= 0X3F 

#if LCD_D4_OUT_REG == 0 || LCD_D5_OUT_REG == 0 || LCD_D6_OUT_REG == 0 || LCD_D7_OUT_REG == 0 || \
    LCD_RS_OUT_REG == 0 || LCD_E_OUT_REG == 0
#error "SOME OF THE LCD PINS ARE INPUTS ONLY"
#endif

#if (LCD_D4_DDR_REG == 0 || LCD_D5_DDR_REG == 0 || LCD_D6_DDR_REG == 0 || LCD_D7_DDR_REG == 0)  \
     &&  LCD_READ_REQUIRED == 1 
#error "THE LCD DATA PINS ARE NOT BIDIRECTIONAL"
#endif

#endif  /* #if SREG > 0X3F */ 

#endif  /* #if defined(__AVR_ATmega103__) */

#define LCD_DONT_READ_BUSY_FLAG  0
#define NUMBER_OF_LCD_UNITS      1

/*######################################################################################################*/
#elif LCD_IO_MODE == 7
/*######################################################################################################*/

#define lcd_toggle_e()   ({ cbivar(lcd_E_port, lcd_E_pin); LCD_E_DELAY(); \
                            sbivar(lcd_E_port, lcd_E_pin); LCD_E_DELAY(); \
                         })

#if SREG > 0X3F
#define IO_TO_MEM_OFFSET  0
#else
#define IO_TO_MEM_OFFSET  0X20
#endif

#define LCD_READ_REQUIRED 0  

#if LCD_0_LINES == 1 || LCD_0_LINES == 2 || LCD_0_LINES == 4
#else
#error THE LCD LINES MUST BE 1 OR 2 OR 4 !
#endif

#if LCD_1_LINES == 1 || LCD_1_LINES == 2 || LCD_1_LINES == 4
#else
#error THE LCD LINES MUST BE 1 OR 2 OR 4 !
#endif

#if NUMBER_OF_LCD_UNITS >= 3
#if LCD_2_LINES == 1 || LCD_2_LINES == 2 || LCD_2_LINES == 4
#else
#error THE LCD LINES MUST BE 1 OR 2 OR 4 !
#endif
#endif


#define LCD_D4_OUT_REG        CONCAT1(PORT, LCD_DATA4_PORT)
#define LCD_D5_OUT_REG        CONCAT1(PORT, LCD_DATA5_PORT)
#define LCD_D6_OUT_REG        CONCAT1(PORT, LCD_DATA6_PORT)
#define LCD_D7_OUT_REG        CONCAT1(PORT, LCD_DATA7_PORT)
#define LCD_RS_OUT_REG        CONCAT1(PORT, LCD_RS_SIGNAL_PORT)
#define LCD_0_E_OUT_REG       CONCAT1(PORT, LCD_0_E_SIGNAL_PORT)
#define LCD_1_E_OUT_REG       CONCAT1(PORT, LCD_1_E_SIGNAL_PORT)

#if NUMBER_OF_LCD_UNITS >=3
#define LCD_2_E_OUT_REG       CONCAT1(PORT, LCD_2_E_SIGNAL_PORT)
#endif

#define LCD_D4_DDR_REG        CONCAT1(DDR, LCD_DATA4_PORT)
#define LCD_D5_DDR_REG        CONCAT1(DDR, LCD_DATA5_PORT)
#define LCD_D6_DDR_REG        CONCAT1(DDR, LCD_DATA6_PORT)
#define LCD_D7_DDR_REG        CONCAT1(DDR, LCD_DATA7_PORT)
#define LCD_RS_DDR_REG        CONCAT1(DDR, LCD_RS_SIGNAL_PORT)
#define LCD_0_E_DDR_REG       CONCAT1(DDR, LCD_0_E_SIGNAL_PORT)
#define LCD_1_E_DDR_REG       CONCAT1(DDR, LCD_1_E_SIGNAL_PORT)

#if NUMBER_OF_LCD_UNITS >=3
#define LCD_2_E_DDR_REG       CONCAT1(DDR, LCD_2_E_SIGNAL_PORT)
#endif

#if defined(__AVR_ATmega103__)
#if SREG > 0X3F

#if LCD_D4_OUT_REG == 0x20 || LCD_D5_OUT_REG == 0x20 || LCD_D6_OUT_REG == 0x20 ||  \
    LCD_D7_OUT_REG == 0x20 || LCD_RS_OUT_REG == 0x20 || LCD_0_E_OUT_REG == 0x20 || \
    LCD_1_E_OUT_REG == 0x20
#error "SOME OF THE LCD PINS ARE INPUTS ONLY"
#endif

#if NUMBER_OF_LCD_UNITS >=3 &&  LCD_2_E_OUT_REG == 0x20
#error "THE LCD LCD_2_E_PIN IS INPUT ONLY"
#endif


#elif SREG <= 0X3F 

#if LCD_D4_OUT_REG == 0 || LCD_D5_OUT_REG == 0 || LCD_D6_OUT_REG == 0 || LCD_D7_OUT_REG == 0 || \
    LCD_RS_OUT_REG == 0 || LCD_0_E_OUT_REG == 0 || LCD_1_E_OUT_REG == 0
#error "SOME OF THE LCD PINS ARE INPUTS ONLY"
#endif

#if NUMBER_OF_LCD_UNITS >=3 &&  LCD_2_E_OUT_REG == 0
#error "THE LCD LCD_2_E_PIN IS INPUT ONLY"
#endif

#endif  /* #if SREG > 0X3F */ 

#endif  /* #if defined(__AVR_ATmega103__) */

#define LCD_DONT_READ_BUSY_FLAG  0
#define LCD_READ_REQUIRED        0

/*######################################################################################################*/
#endif /* elif LCD_IO_MODE == 7 */ 
/*######################################################################################################*/

/* lcd_write() mode argument values */
#define INIT_MODE                2
#define DATA_MODE                1
#define CMD_MODE                 0

#if LCD_READ_REQUIRED == 1  
/* lcd_read() argument values */
#define BUSY_FLAG                0
#define ADDRESS_COUNTER          0
#define DATA                     1

#endif



/*######################################################################################################*/
/*                                        TYPE DEFINITIONS                                              */
/*######################################################################################################*/


typedef unsigned char  u08;
typedef unsigned short u16;


extern unsigned char x,y;

/*######################################################################################################*/

#if LCD_CLRLINE_NEEDED == 1
void lcd_clrline(unsigned char line)
/* clear a specific lcd line and set cursor at the start of the line */
{
#if LCD_IO_MODE == 7 || LCD_IO_MODE == 3
   if(lcd_lines==1)
    {
       lcd_gotoxy(0,0);
       while(x<lcd_chars_per_line) { lcd_putc(' '); }
       lcd_gotoxy(0,0);
    }    
   else if(line<lcd_lines)
         {
            lcd_gotoxy(0,line);
            while(x<lcd_chars_per_line) { lcd_putc(' '); }
            lcd_gotoxy(0,line);
         }   

#elif LCD_IO_MODE == 6 || LCD_IO_MODE == 2 

#if LCD_LINES == 1
  lcd_gotoxy(0,0);
  while(x<LCD_CHARS_PER_LINE) { lcd_putc(' '); }
  lcd_gotoxy(0,0);
#elif (LCD_LINES == 2 || LCD_LINES == 4)  
  if(line<LCD_LINES)
   {
      lcd_gotoxy(0,line);
      while(x<LCD_CHARS_PER_LINE) { lcd_putc(' '); }
      lcd_gotoxy(0,line);
   }   
#endif  /*  #if LCD_LINES == 1 -> #elif (LCD_LINES == 2 || LCD_LINES == 4) */

#endif  /*  #if LCD_IO_MODE == 7 -> #elif LCD_IO_MODE == 6 || LCD_IO_MODE == 2 */
}
#endif   /* #if LCD_CLRLINE_NEEDED == 1 */
