/* vi:set et ts=4 sw=4 ai ft=c ff=dos: */
#include "rtclib.h"

static uint8_t rtc_enable_osc ( void ) ;

uint8_t
rtc_enable_osc ( void )
{
    uint8_t     rc ;
    uint8_t     sec = 0;

    rc = twi_send_start ( ) ;
    if ( rc == SUCCESS )               
        rc = twi_send_slave_address( DS1307_ADDR + TW_WRITE ) ;
    if ( rc == SUCCESS )
        rc = twi_send_byte ( 0 ) ;

    if ( rc == SUCCESS )
        rc = twi_send_start ( ) ;
    if ( rc == SUCCESS )
        rc = twi_send_slave_address( DS1307_ADDR + TW_READ ) ;
    if ( rc == SUCCESS )
        sec = twi_receive_byte( 0 ) & 0x7F ;

    if ( rc == SUCCESS )
        rc = twi_send_start ( ) ;
    if ( rc == SUCCESS )               
        rc = twi_send_slave_address
            ( DS1307_ADDR + TW_WRITE ) ;
    if ( rc == SUCCESS )
    {
        rc = twi_send_byte ( 0 ) ;
        if ( rc == SUCCESS )
            rc = twi_send_byte ( sec ) ;
        twi_send_stop ( ) ;
    }
    return rc ;
}

uint8_t rtc_read_time ( cTime_t *tm )
{
    uint8_t     rc ;

    rc = twi_send_start ( ) ;
    if ( rc == SUCCESS )               
        rc = twi_send_slave_address ( DS1307_ADDR + TW_WRITE ) ;
    if ( rc == SUCCESS )               
        rc = twi_send_byte ( 0x00 ) ;

    if ( rc == SUCCESS )
        rc = twi_send_start ( ) ;
    if ( rc == SUCCESS )
        rc = twi_send_slave_address( DS1307_ADDR + TW_READ ) ;

    // Reads one byte from the I2C device, request more data from device
    // and write into structure variables
    if ( rc == SUCCESS )
    {
        tm->sec = bcd_to_bin ( twi_receive_byte ( 1 ) ) ;
        tm->min = bcd_to_bin ( twi_receive_byte ( 1 )  & 0x7f ) ;
        tm->hour = bcd_to_bin ( twi_receive_byte ( 1 ) & 0x3f ) ;
        tm->day = bcd_to_bin ( twi_receive_byte ( 1 ) ) ;
        tm->date = bcd_to_bin ( twi_receive_byte ( 1 ) & 0x3f ) ;
        tm->month = bcd_to_bin ( twi_receive_byte ( 1 ) & 0x1f ) ;
        tm->year = bcd_to_bin( twi_receive_byte ( 0 ) & 0xFF ) ;
    
        twi_send_stop ( ) ;
    }
    return 1 ;
}

void
WriteTimeToLCD ( cTime_t *time )
{
    uint8_t     TimeBuffer[16];
    uint8_t     DateBuffer[16];
    static char *days[ ] = {
        "Mon", "Tue", "Wed",
        "Thu", "Fri", "Sat", "Sun"
    } ;

    sprintf( TimeBuffer, "%3s %02d:%02d:%02d\n",
                time->day <= 7 ? days [ time->day -1 ] : " ",
                time->hour, time->min, time->sec ) ;
    sprintf ( DateBuffer, "    %02d-%02d-%02d",
        time->date, time->month, time->year ) ;

    lcd_puts(TimeBuffer);
    lcd_puts(DateBuffer);
}

void
rtc_display_time ( void )
{
    cTime_t TimeStamp ;

    for ( ; ; )
    {
        rtc_read_time ( &TimeStamp ) ;
        lcd_gotoxy ( 0, 0 ) ;
        WriteTimeToLCD ( &TimeStamp ) ;
        _delay_ms( 50 ) ;
    }
}

void
rtc_init ( void )
{
    twi_init ( ) ;
    rtc_enable_osc ( ) ;
}
