/* vi:set ts=4 sw=4 ai et ff=dos ft=c: */
#include "twi_driver.h"

/* Address of ATMega32 when used as 'slave receiver' */
#ifndef OWN_ADDR
#define OWN_ADDR     0x60
#endif

static uint8_t _get_byte ( uint8_t*, char not_last_byte ) ;
static void twi_wait_int ( void ) ;

void twi_init ( void )
{
    TWAR = OWN_ADDR ;
    TWBR = (F_CPU / 100000UL - 16) / 2 ;
    TWCR = _BV( TWEN ) ;
    return ;
}

void twi_wait_int ( void )
{
    while ( ! ( TWCR & _BV( TWINT ) ) )
        ;
}    

uint8_t twi_send_start ( void )
{
    TWCR = _BV ( TWINT ) | _BV ( TWSTA ) | _BV ( TWEN ) ;

    twi_wait_int ( ) ;  // Wait for TWI interrupt flag set
    if ( (TWSR != TW_START) && (TWSR != TW_REP_START) )
        return TWSR ;
    return SUCCESS ;
}

void
twi_send_stop(void)
{
    TWCR = _BV ( TWINT ) | _BV ( TWSTO ) | _BV ( TWEN ) ;
}

uint8_t
twi_send_byte( uint8_t data )
{
    twi_wait_int();

    TWDR = data;
    TWCR = _BV ( TWINT ) | _BV ( TWEN ) ;

    twi_wait_int ( ) ;

    if ( TWSR != TW_MT_DATA_ACK )
        return TWSR ;
    return SUCCESS ;
}

/****************************************************************************
    Send a SLA + W/R to the bus
****************************************************************************/
uint8_t
twi_send_slave_address ( uint8_t adr )
{
    twi_wait_int ( ) ;

    TWDR = adr ;
    TWCR = _BV ( TWINT ) | _BV ( TWEN ) ;

    twi_wait_int ( ) ;

    if ( TWSR != TW_MT_SLA_ACK && TWSR != TW_MR_SLA_ACK )
        return TWSR ;
    return SUCCESS ;
}

uint8_t _get_byte( uint8_t *rx_ptr,char not_last_byte )
{
    twi_wait_int ( ) ;

    /* When receiving the last byte from the slave it will be sent a NACK to 
       make the slave stop transmitting, all bits before the last will get a
       ACK */
    if (not_last_byte)                           //Not the last byte
        // Clear int flag and enable acknowledge to receive data.
        TWCR = _BV ( TWINT ) | _BV ( TWEN ) | _BV ( TWEA ) ;
    else                                    //Last byte
        /* Clear int flag to and do not enable acknowledge to tell the slave 
           to stop transmitting */
        TWCR = _BV ( TWINT ) | _BV ( TWEN ) ;
    twi_wait_int ( ) ;                         // Wait for TWI interrupt flag set

    *rx_ptr = TWDR ;                         // Save received byte

    /* If ACK has been transmitted or this was the last byte and NACK has been
       sent -> return SUCCESS, else return TWSR */
    if ( ( ( TWSR == TW_MR_DATA_NACK ) && ( not_last_byte == 0 ) ) || 
			(TWSR == TW_MR_DATA_ACK) )
        return SUCCESS ;   
    return TWSR ;
}

uint8_t twi_receive_byte ( char not_last_byte )
{
    uint8_t   ch ;
    uint8_t   status = _get_byte ( & ch , not_last_byte ) ;
    if ( status == SUCCESS )
        return ch ;
    else
        return 0xff ;
}
