# vi:set ft=make noet ai ts=4 sw=4:
#
# Written by Chris Efstathiou 
# Date 20/Dec/2002
# Updated by Asang Dani
# Date 31/Aug/2007

#---------------- Variables used throughout the makefile ----------------

CC		= avr-gcc
AS		= avr-gcc -x assembler-with-cpp	
AR		= avr-ar
ARFLAGS	= cru
REMOVE	= rm -f
REMOVEDIR = rm -rf
RN		= mv
WINSHELL = cmd
OBJCOPY	= avr-objcopy
OBJDUMP	= avr-objdump
FORMAT	= ihex
SIZE 	= avr-size -A
LIBDIR	= $(TOPDIR)/lib
CTAGS	= ctags
CTAGSF	=--file-tags=yes -R
#GCCDIR	= $(shell cygpath -m `which avr-gcc.exe` | sed -e 's;/bin/avr-gcc.exe;;g')/avr/include
#
# Change this to suit your local installation
#
GCCDIR	= C:/WinAVR/avr/include
OBJDIR_BASE = obj
OBJDIR = $(OBJDIR_BASE)/$(MCU)

LCDDIR = $(TOPDIR)/lib/lcd
MENUDIR = $(TOPDIR)/lib/menu
BUTTONDIR = $(TOPDIR)/lib/buttons
RPRINTFDIR = $(TOPDIR)/lib/rprintf
CMDLINEDIR = $(TOPDIR)/lib/cmdline
UARTDIR  = $(TOPDIR)/lib/uart
TERMIODIR = $(TOPDIR)/lib/termio
RTCDIR = $(TOPDIR)/lib/rtclib

TAGDIRS  = $(GCCDIR) $(LCDDIR) $(I2DIR) $(RPRINTFDIR)
TAGDIRS += $(CMDLINEDIR) $(UARTDIR) $(TOPDIR)/inc
TAGDIRS += $(BUTTONDIR) $(MENUDIR)
TAGDIRS += $(TERMIODIR)
TAGDIRS += $(RTCDIR)

#---------------- Programming Options (avrdude) ----------------

# Programming hardware: alf avr910 avrisp bascom bsd 
# dt006 pavr picoweb pony-stk200 sp12 stk200 stk500
#
# Type: avrdude -c ?
# to get a full listing.
#
AVRDUDE = avrdude
AVRDUDE_PROGRAMMER = pony-stk200

# com1 = serial port. Use lpt1 to connect to parallel port.
AVRDUDE_PORT = lpt1

AVRDUDE_WRITE_FLASH = -u -U flash:w:$(TARGET).hex
AVRDUDE_WRITE_EEPROM = -u -U eeprom:w:$(TARGET).eep

# Uncomment the following if you want avrdude's erase cycle counter.
# Note that this counter needs to be initialized first using -Yn,
# see avrdude manual.
#AVRDUDE_ERASE_COUNTER = -y

# Uncomment the following if you do /not/ wish a verification to be
# performed after programming the device.
#AVRDUDE_NO_VERIFY = -V

# Increase verbosity level.  Please use this when submitting bug
# reports about avrdude. See <http://savannah.nongnu.org/projects/avrdude> 
# to submit bug reports.
#AVRDUDE_VERBOSE = -v -v

AVRDUDE_FLAGS = -p $(MCU) -P $(AVRDUDE_PORT) -c $(AVRDUDE_PROGRAMMER)
AVRDUDE_FLAGS += $(AVRDUDE_NO_VERIFY)
AVRDUDE_FLAGS += $(AVRDUDE_VERBOSE)
AVRDUDE_FLAGS += $(AVRDUDE_ERASE_COUNTER)

TARGET_FLASH_READ_FILE = `cygpath -w $(PWD)/$(TARGET)_flash.bin`
TARGET_EE_READ_FILE = `cygpath -w $(PWD)/$(TARGET)_eep.bin`

AVRDUDE_READ_FLASH = -U flash:r:"$(TARGET_FLASH_READ_FILE)":r
AVRDUDE_READ_EEPROM = -U eeprom:r:"$(TARGET_EE_READ_FILE)":r

#---------------- Debugging Options ----------------

# For simulavr only - target MCU frequency.
DEBUG_MFREQ = $(F_CPU)

# Set the DEBUG_UI to either gdb or insight.
# DEBUG_UI = gdb
DEBUG_UI = insight

# Set the debugging back-end to either avarice, simulavr.
# DEBUG_BACKEND = avarice
DEBUG_BACKEND = simulavr

# GDB Init Filename.
GDBINIT_FILE = __avr_gdbinit

# When using avarice settings for the JTAG
JTAG_DEV = /dev/com1

# Debugging port used to communicate between GDB / avarice / simulavr.
DEBUG_PORT = 4242

# Debugging host used to communicate between GDB / avarice / simulavr, normally
#     just set to localhost unless doing some sort of crazy debugging when 
#     avarice is running on a different computer.
DEBUG_HOST = localhost


#============================================================================

################################################################################
#           U S E R   C O N F I G U R A T I O N  B L O C K  S T A R T
################################################################################

# Put AVRGCC version (gcc --version) here (needed for finding the correct
# header files)
AVRGCC_VERSION	= 345

# Put the MCU clock Frequency in Hertz here (This setting is passed to all files)
F_CPU	= 8000000

# Put name of target mcu here
# (at90s8515, at90s8535, attiny22, atmega603 etc.)
ifndef MCU
MCU = atmega32
endif

INCDIR += $(TOPDIR)/inc
INCDIR += .

ifndef SOURCES
SOURCES = $(wildcard *.c)
endif

ifdef USE_LCD
LIB_SOURCES += lcd_io.c lcd_io_optional.c
VPATH += $(LCDDIR)
endif

ifdef USE_RPRINTF
LIB_SOURCES += rprintf.c vt100.c buffer.c
VPATH += $(RPRINTFDIR)
USE_UART=yes
endif

ifdef USE_UART
LIB_SOURCES += uart.c
VPATH += $(UARTDIR)
endif

ifdef USE_CMDLINE
LIB_SOURCES += cmdline.c
VPATH += $(CMDLINEDIR)
endif

ifdef USE_MENU
LIB_SOURCES += menu.c
VPATH += $(MENUDIR)
endif

ifdef USE_BUTTONS
LIB_SOURCES += buttons.c
VPATH += $(BUTTONDIR)
endif

ifdef USE_TERMIO
LIB_SOURCES += termio.c
VPATH += $(TERMIODIR)
endif

ifdef USE_RTCLIB
LIB_SOURCES += rtclib.c twi_driver.c
VPATH += $(RTCDIR)
endif

MSG_BEGIN = -------- begin --------
MSG_END = --------  end  --------
MSG_SIZE_BEFORE = Size before: 
MSG_SIZE_AFTER = Size after:
MSG_LINKING = Linking:
MSG_CLEANING = Cleaning project:
MSG_ASSEMBLING = Assembling
MSG_COMPILING = Compiling
MSG_FLASH = Creating load file for Flash:
MSG_EEPROM = Creating load file for EEPROM:
MSG_READ_EE_FLASH = Reading contents of Flash and EEPROM
MSG_ASM = Creating disassembly file:
LIB=$(LIBDIR)/libkset$(MCU).a

ifdef NOLIB
LIB=
endif

INCLUDES = $(patsubst %,-I%,$(INCDIR))

###### output format can be srec, ihex (avrobj is always created) #######

################################################################################
#           U S E R   C O N F I G U R A T I O N  B L O C K  E N D
################################################################################

########## Normally you shouldn't need to change anything in the below lines ###

CFLAGS += -g -Os -funsigned-char -funsigned-bitfields -fpack-struct -fshort-enums -Wall -Wstrict-prototypes -Wa,-ahlms=$(<F:%.c=$(OBJDIR)/%.lst)

ASFLAGS = -g -Os -funsigned-char -funsigned-bitfields -fpack-struct -fshort-enums -Wall -Wstrict-prototypes -Wa,-ahlms=$(<F:%.S=$(OBJDIR)/%.lst)

ifdef USE_DEBUGGING
CFLAGS +=-g -gdwarf-2
ASFLAGS += -Wa,-gstabs
endif

# Linker flags
LDFLAGS = -Wl,-Map=$(OBJDIR)/$(TARGET).map
LDFLAGS += -Wl,--cref

# Define all project specific object files
OBJ	= $(ASRC:%.S=$(OBJDIR)/%.o) $(SOURCES:%.c=$(OBJDIR)/%.o) 
CFLAGS += -mmcu=$(MCU)
ASFLAGS += -mmcu=$(MCU)
LDFLAGS += -mmcu=$(MCU)

# Pass the MPU clock frequency and version info to flags
ASFLAGS += -DAVRGCC_VERSION=$(AVRGCC_VERSION)
CFLAGS += -DAVRGCC_VERSION=$(AVRGCC_VERSION)
ASFLAGS += -DF_CPU=$(F_CPU) $(INCLUDES)
CFLAGS += -DF_CPU=$(F_CPU)
 
# Create object files directory
$(shell mkdir -p $(OBJDIR) 2>/dev/null)

# Compiler flags to generate dependency files.
GENDEPFLAGS = -MD -MP -MF .dep/$(@F).d

CFLAGS += $(INCLUDES)
ASFLAGS += $(INCLUDES)
CFLAGS += $(GENDEPFLAGS)

# This defines the aims of the make process

ifeq ($(TARGET),libkset$(MCU).a)

LIBKSET_OBJS = $(LIB_ASRC:%.S=$(OBJDIR)/%.o) $(LIB_SOURCES:%.c=$(OBJDIR)/%.o)

all: lib/$(TARGET)

else

all: begin sizebefore elf hex eep sizeafter asm end

endif

elf: $(OBJDIR)/$(TARGET).elf
hex: $(TARGET).hex
eep: $(TARGET).eep
asm: $(OBJDIR)/$(TARGET).asm

begin:
	@echo $(MSG_BEGIN)

end:
	@echo
	@echo $(MSG_END)

HEXSIZE = $(SIZE) --target=$(FORMAT) $(TARGET).hex
ELFSIZE = $(SIZE) -B $(OBJDIR)/$(TARGET).elf

sizebefore:
	@if test -f $(OBJDIR)/$(TARGET).elf; then echo; echo $(MSG_SIZE_BEFORE); $(ELFSIZE); \
	echo; fi

sizeafter:
	@if test -f $(OBJDIR)/$(TARGET).elf; then echo; echo $(MSG_SIZE_AFTER); $(ELFSIZE); \
	$(AVRMEM) 2>/dev/null; echo; fi

program: $(TARGET).hex $(TARGET).eep
	$(AVRDUDE) $(AVRDUDE_FLAGS) $(AVRDUDE_WRITE_FLASH) $(AVRDUDE_WRITE_EEPROM)

read: begin read_real end

read_real: begin
	@echo $(MSG_READ_EE_FLASH)
	@echo
	$(AVRDUDE) $(AVRDUDE_FLAGS) $(AVRDUDE_READ_FLASH) $(AVRDUDE_READ_EEPROM)

# Generate avr-gdb config/init file which does the following:
#     define the reset signal, load the target file, connect to target, and set 
#     a breakpoint at main().
gdb-config: 
	@$(REMOVE) $(GDBINIT_FILE)
	@echo define reset >> $(GDBINIT_FILE)
	@echo SIGNAL SIGHUP >> $(GDBINIT_FILE)
	@echo end >> $(GDBINIT_FILE)
	@echo file $(TARGET).elf >> $(GDBINIT_FILE)
	@echo target remote $(DEBUG_HOST):$(DEBUG_PORT)  >> $(GDBINIT_FILE)
ifeq ($(DEBUG_BACKEND),simulavr)
	@echo load  >> $(GDBINIT_FILE)
endif
	@echo break main >> $(GDBINIT_FILE)

debug: gdb-config $(OBJDIR)/$(TARGET).elf
ifeq ($(DEBUG_BACKEND), avarice)
	@echo Starting AVaRICE - Press enter when "waiting to connect" message displays.
	@$(WINSHELL) /c start avarice --jtag $(JTAG_DEV) --erase --program --file \
	$(TARGET).elf $(DEBUG_HOST):$(DEBUG_PORT)
	@$(WINSHELL) /c pause

else
	@$(WINSHELL) /c start simulavr --gdbserver --device $(MCU) --clock-freq \
	$(DEBUG_MFREQ) --port $(DEBUG_PORT)
endif
	@$(WINSHELL) /c start avr-$(DEBUG_UI) --command=$(GDBINIT_FILE)

# Compile: create object files from C source files.
$(OBJDIR)/%.o : %.c
	@echo
	@echo $(MSG_COMPILING) $<
	$(CC) -c $(CFLAGS) $< -o $@ 

# Assemble: create object files from assembler source files.
$(OBJDIR)/%.o : %.S
	@echo
	@echo $(MSG_ASSEMBLING) $<
	$(CC) -c $(ASFLAGS) $< -o $@

%s : %c
	$(CC) -S $(CFLAGS) $< -o $@

# Link: instructions to create elf output file from object files
$(OBJDIR)/%.elf: $(OBJ)
	@echo
	@echo $(MSG_LINKING) $@
	$(CC) $(OBJ) $(LIB) $(LDFLAGS) -o $@

# Create bin (ihex, srec) file from elf output file
%hex: $(OBJDIR)/%elf
	@echo
	@echo $(MSG_FLASH) $@
	$(OBJCOPY) -O $(FORMAT) -R .eeprom $< $@

$(OBJDIR)/$(TARGET).asm: $(OBJDIR)/$(TARGET).elf
	@echo
	@echo $(MSG_ASM) $@
	$(OBJDUMP) -h -S $< > $@

%.eep: $(OBJDIR)/%.elf
	@echo
	@echo $(MSG_EEPROM) $@
	$(OBJCOPY) -j .eeprom --set-section-flags=.eeprom="alloc,load" --change-section-lma .eeprom=0 -O $(FORMAT) $< $@

clean: begin clean_real end

ifeq ($(TARGET),libkset$(MCU).a)

clean_real:
	@echo
	@echo $(MSG_CLEANING)
	$(REMOVEDIR) $(OBJDIR)
	$(REMOVEDIR) .dep
	$(REMOVE) tags cscope.files cscope.out

else

clean_real:
	@echo
	@echo $(MSG_CLEANING)
	$(REMOVEDIR) $(OBJDIR_BASE)
	$(REMOVE) $(TARGET).hex
	$(REMOVEDIR) .dep
	$(REMOVE) $(TARGET).eep
	$(REMOVE) tags cscope.files cscope.out
	$(REMOVE) $(GDBINIT_FILE)
	$(REMOVE) $(TARGET_FLASH_READ_FILE)
	$(REMOVE) $(TARGET_EE_READ_FILE)

endif

size:
	$(SIZE) $(OBJDIR)/$(TARGET).elf

tags:
	/bin/find . -name "*.[CchHsS]" > cscope.files
	/bin/find $(TAGDIRS) -name "*.[cChHsS]" >> cscope.files
	/bin/find $(VPATH) -name "*.[cChHsS]" >> cscope.files
	$(CTAGS) $(CTAGSF) -L cscope.files

# Include the dependency files.
-include $(shell mkdir .dep 2>/dev/null) $(wildcard .dep/*)

.PHONY : all build elf hex lss asm coff extcoff \
clean clean_list program debug gdb-config

.PRECIOUS : $(OBJ)

# "C:\WinAVR\bin\avrdude.exe" -p m16 -c pony-stk200 -P lpt1 -U flash:w:"G:\MyDocuments\Programs\myavr\led_blinking_mmio\led_blinking_mmio.hex":i -U flash:v:"G:\MyDocuments\Programs\myavr\led_blinking_mmio\led_blinking_mmio.hex":i
