# vim:set ft=make noet ai ts=4 sw=4:

RM	= rm
MV	= mv
INCDIR= .
CTAGS= ctags
CTAGSF=--file-tags=yes -R
OBJDIR = obj
GENDEPFLAGS = -MD -MP -MF .dep/$(@F).d
TAGDIRS=c:/cygwin/usr/include

TARGETSRC=$(wildcard $(TARGET).c*)

ifndef SOURCES
SOURCES = $(wildcard *.c)

ifneq "$(wildcard *.cpp)" ""
USE_CPP=1
CXX = g++
SOURCES += $(wildcard *.cpp)
endif

endif

#additional libraries and object files to link
LIB     =

ifndef INCLUDES
INCLUDES = 
endif

CFLAGS += -Wall
CFLAGS += $(GENDEPFLAGS)

ifdef USE_DEBUGGING
CFLAGS +=-g
endif

CFLAGS += $(patsubst %,-I%,$(INCLUDES))

#define all project specific object files
OBJ	= $(SOURCES:%.c=$(OBJDIR)/%.o) 

ifndef USE_CPP
CXX	= gcc
else
CXX	= g++
endif

.DEFAULT: all
.PHONY: all

all: $(OBJDIR)/$(TARGET).exe

$(OBJDIR)/%.o: %.c 
	@if [ ! -d $(OBJDIR) ]; then mkdir $(OBJDIR); fi
	$(CXX) -c $(CFLAGS) $< -o $@

$(OBJDIR)/%.exe: $(OBJ)
	$(CXX) $(OBJ) $(LIB) $(LDFLAGS) -o $@

depend: $(SOURCES)
	@echo "Generating dependencies for $<"
	@$(CXX) -MM $(SOURCES) | sed -e 's;^\(.*\)\.o:;$(OBJDIR)/\1.o:;g' > .depend

.PHONY: clean
clean:
	-$(RM) -rf $(OBJDIR)

reallyclean: clean
	-$(RM) tags .depend

tags:
	/bin/find . -name "*.[CchHsS]" > cscope.files
	/bin/find $(TAGDIRS) -name "*.[cChHsS]" >> cscope.files
	$(CTAGS) $(CTAGSF) -L cscope.files

# Include the dependency files.
-include $(shell mkdir .dep 2>/dev/null) $(wildcard .dep/*)

.PRECIOUS: $(OBJ)
